# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
from itertools import chain

from formed.meta.data import RelevantVisitor, RuleLeaf, ExternalChoiceListLeaf,\
     ConditionalNode


def conditional_parents(nc):
    """Return a list with all ConditionalNode ancestors of NodeComponen nc"""
    # ignore the last item of the return value of nc.pathToRoot()
    # because that is nc itself.
    return [item for item in nc.pathToRoot()[:-1]
                 if isinstance(item, ConditionalNode)]



class Document(object):

    def __init__(self, root=None):
        if root is None:
            self.root = RootNode()
        else:
            self.root = root
        self.root.document = self
        self.name2nc = None

    def init(self):
        self.buildIndices()
        self.validate_tree()

    def buildIndices(self):
        name2nc = {}
        name2rules = {}
        name2markingrules = {}
        name2conditionals = {}

        def callback(nc):
            if isinstance(nc, RuleLeaf):
                expr = nc.getExpr()
                for var in expr.getDependencies():
                    name2rules.setdefault(var, []).append(nc)
                marked_name = nc.getMarkedName()
                if marked_name:
                    name2markingrules.setdefault(marked_name, []).append(nc)

            name = nc.getName()
            if name:
                if name in name2nc:
                    print >> sys.stderr, ("WARNING: name %r used multiple times"
                                          % name)
                name2nc[name] = nc
                name2conditionals[name] = conditional_parents(name2nc[name])
            else:
                print >> sys.stderr, "WARNING: ignore unnamed: %s" % repr(nc)

        visitor = RelevantVisitor(callback)
        self.visit(visitor.visitor)

        self.name2nc = name2nc
        self.name2rules = name2rules
        self.name2markingrules = name2markingrules
        self.name2conditionals = name2conditionals

    def validate_tree(self):
        """Check the tree for inconsistencies and print warnings.
        Currently this checks whether all dependencies of expressions in
        rules or conditionals are actually defined in the tree.
        """
        for item in self.findAllByClass(RuleLeaf):
            self.validate_expr_dependencies(item)
        for item in self.findAllByClass(ConditionalNode):
            self.validate_expr_dependencies(item)
            self.validate_conditional_dependencies(item)
        self.validate_external_choice_leafs()

    def validate_expr_dependencies(self, node):
        """Check whether all dependencies of the expression in node exist.
        The node must be have a getExpr method that returns an Expr object.
        For each dependency that doesn't exist in the formed tree, a
        message is printed to stderr.
        """
        for dependency in node.getExpr().getDependencies():
            if dependency not in self.name2nc:
                if isinstance(node, RuleLeaf):
                    kind = "rule"
                else:
                    kind = "conditional"
                print >>sys.stderr, ("WARNING: %s %r has unknown"
                                     " dependency %r"
                                     % (kind, node.getName(), dependency))

    def validate_conditional_dependencies(self, conditional):
        """Checks that the conditional does not depend on other conditionals.

        A conditional depends on other conditionals if its expression
        references fields that belong to other conditionals.  This is
        not yet supported fully by libformed.  There are basically three
        cases to consider:

         1) The other conditional is a parent of the conditional whose
            expression is being verified.  This case is supported, so no
            message is printed.

         2) The other conditional is in fact the conditional whose
            expression is being verified or one of its children.  This
            case is most likely an error, but it is actually
            occasionally useful to specify such conditionals, so we
            print a warning.

            An example for a use-case is a conditional that true only if
            one of the fields in the conditionals is set.  This can be
            used to forbid setting the field through normal means via
            the generated forms.  This field may still be set through
            some other special means.

         3) The other conditional is not related to the conditional
            whose expression is being verified.  This should be OK in
            principle, but this is not yet properly supported by
            libformed, so a message is printed to stderr.
        """
        parents = conditional_parents(conditional)
        for name in conditional.getExpr().getDependencies():
            for cond in self.name2conditionals[name]:
                if cond is conditional:
                    # this is case 2) in the doc-string.  Likely an error,
                    # but since it's sometimes useful, only print a warning.
                    print >>sys.stderr, ("Conditional %r refers to itself via"
                                         " %r. This may be an error."
                                         % (conditional.getName(), name))
                elif cond in parents:
                    # this is case 1) in the doc-string.  This is OK
                    continue
                else:
                    # otherwise, this is case 3) in doc-string.  Print a
                    # warning.
                    print >>sys.stderr, ("Conditional %r refers to other"
                                         " conditional %r. This is not yet"
                                         " supported in libformed"
                                         % (conditional.getName(),
                                            cond.getName()))

    def validate_external_choice_leafs(self):
        """Force loading of external choice data.
        If the file referenced by an ExternalChoiceListLeaf instance
        cannot be loaded an exception will be raised.
        """
        for external in self.findAllByClass(ExternalChoiceListLeaf):
            # calling getChildren forces loading of the external choice
            # data.
            external.getChildren()

    def getVersion(self):
        return self.root.getVersion()

    def getRoot(self):
        return self.root

    def importDocument(self, ndocument):
        root = self.root
        nroot = ndocument.root
        if not root or not nroot:
            return

        for child in nroot.children[:]:
            nroot.removeChild(child)
            root.addChild(child)

    def getNames(self):
        names = set()
        root = self.root
        if not root:
            return names
        for nc in root.walk():
            name = nc.getName()
            if name:
                names.add(name)
        return names

    def findByName(self, name, linear_lookup=False):
        if self.name2nc:
            try:
                return self.name2nc[name]
            except KeyError:
                if not linear_lookup:
                    return None

        if self.root:
            for nc in self.root.walk():
                if name == nc.getName():
                    return nc
        return None

    def findByClassAndName(self, clazz, name=None):
        if self.root:
            for nc in self.root.walk():
                if isinstance(nc, clazz) and (not name or nc.getName() == name):
                    return nc
        return None

    def findAllByClass(self, clazz):
        if self.root:
            for nc in self.root.walk():
                if isinstance(nc, clazz):
                    yield nc

    def visit(self, visitor):
        if self.root:
            self.root.visit(visitor)

    def walk(self):
        if self.root:
            for nc in self.root.walk():
                yield nc

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8 :
