# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

from sys import maxint
import re
from datetime import date

import data as data
from formed.meta import UNKNOWN_INT, UNKNOWN_DATE


def ensure_unicode(s):
    """Convert s to a unicode object.
    If it's alredy a unicode object, s is returned as is.  Otherwise
    it's converted to a byte-string which is assumed to be in utf-8 and
    then converted to unicode."""
    if isinstance(s, unicode):
        return s
    return unicode(str(s), 'utf-8')


class ErrorItem:

    SEMANTIC = 0
    RULE = 1

    def __init__(self, type=SEMANTIC, name=None, msg="", bad=None):
        self.type = type
        self.name = name
        self.msg = msg
        self.bad = bad

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return ", ".join(["type: %s" % repr(self.type),
                          "name: %s" % repr(self.name),
                          "msg:  %s" % repr(self.msg),
                          "bad:  %s" % repr(self.bad)])


class SemanticError(Exception):

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return repr(self.value)


class ValueChecker:

    def convert(self, value, nc):
        return value

    def check(self, value, nc):
        return value

    def checkAndConvert(self, value, nc):
        return self.check(self.convert(value, nc), nc)


class TextChecker(ValueChecker):

    def convert(self, value, nc):
        return ensure_unicode(value)

    def check(self, value, nc):
        ml = nc.getMaxLength()
        if ml is not None:
            value = value[:ml]
        return value


class TextAreaChecker(ValueChecker):

    def convert(self, value, nc):
        return ensure_unicode(value)


class IntChecker(ValueChecker):

    def convert(self, value, nc):
        value = value.strip()

        if value[:2].lower() == u"un":
            return UNKNOWN_INT

        try:
            return int(float(value))
        except ValueError:
            raise SemanticError("'%s' ist keine ganze Zahl." % value)

    def check(self, value, nc):
        # UNKNOWN_INT is always valid, even if it's not in the range
        # required by nc
        if value == UNKNOWN_INT:
            return value

        minv = nc.getMinValue()
        if minv is None:
            minv = -maxint - 1
        maxv = nc.getMaxValue()
        if maxv is None:
            maxv = maxint

        minv, maxv = min(minv, maxv), max(minv, maxv)

        if value < minv:
            raise SemanticError(u"%d ist kleiner als der zulässige"
                                u" Minimalwert %d." % (value, minv))

        if value > maxv:
            raise SemanticError(u"%d ist größer als der zulässige"
                                u" Maximalwert %d." % (value, maxv))

        return value

class NumericChecker(ValueChecker):

    def convert(self, value, nc):
        value = value.strip()

        if value[:2].lower() == u"un":
            return float(UNKNOWN_INT)

        try:
            t = value.replace(',', '.')
            return float(t)
        except ValueError:
            raise SemanticError("'%s' ist keine Zahl." % value)

    def check(self, value, nc):
        minv = nc.getMinValue()
        if minv is None:
            minv = -maxint - 1
        maxv = nc.getMaxValue()
        if maxv is None:
            maxv = maxint

        minv, maxv = min(minv, maxv), max(minv, maxv)

        if value < minv:
            raise SemanticError(u"%s ist kleiner als der zulässige"
                                u" Minimalwert %.2f." % (value, minv))

        if value > maxv:
            raise SemanticError(u"%s ist größer als der zulässige"
                                u" Maximalwert %.2f." % (value, maxv))

        return value


def expand_two_digit_year(year):
    """If year was only given with two digits convert it to the intended
    year.  The parameter year should be an int.  If year is less than
    100, the year was specified with only two digits and is actually
    meant to refer to a year in the 20th or 21st centuries.  If the year
    is less than 70 it's assumed to actually refer to 2000 + year,
    otherwise to 1900 + year.
    """
    if year < 100:
        if year < 70:
            year += 2000
        else:
            year += 1900
    return year


class DateChecker(ValueChecker):

    DATE = re.compile(r"(?P<day>[0-9]{1,2})\.(?P<month>[0-9]{1,2})"
                      r"\.(?P<year>[0-9]{2,4})")
    DATE_ISO = re.compile(r"(?P<year>[0-9]{2,4})-(?P<month>[0-9]{1,2})"
                          r"-(?P<day>[0-9]{1,2})")
    DATE_YEAR_ONLY = re.compile(r"(?P<year>[0-9]{4})")

    def convert(self, value, nc):
        value = value.strip()

        if value[:2].lower() == u"un":
            return UNKNOWN_DATE

        for reg in (DateChecker.DATE, DateChecker.DATE_ISO,
                    DateChecker.DATE_YEAR_ONLY):
            m = reg.match(value)
            if not m:
                continue
            try:
                day = int(m.group("day"))
            except IndexError:
                day = 1
            try:
                month = int(m.group("month"))
            except IndexError:
                month = 1
            year = expand_two_digit_year(int(m.group("year")))
            try:
                return date(year, month, day)
            except ValueError:
                raise SemanticError(u"'%s' ist kein gültiges Datum"
                                    u" (Format: TT.MM.JJJJ)." % value)

        raise SemanticError(u"'%s' ist kein gültiges Datum"
                            u" (Format: TT.MM.JJJJ)." % value)


class ChoiceChecker(ValueChecker):

    def convert(self, value, nc):
        return int(value)

    def check(self, value, nc):
        stack = nc.children[:]
        while stack:
            c = stack.pop()
            if isinstance(c, (data.GroupNode, data.ExternalChoiceListLeaf)):
                stack.extend(c.getChildren())
            elif isinstance(c, data.BoolLeaf):
                if value == c.getValue():
                    return value

        raise SemanticError("'%s' is not a valid value for '%s'"
                            % (value, nc.getName()))


class RadioChecker(ValueChecker):

    def convert(self, value, nc):
        return int(value)

    def check(self, value, nc):
        for c in nc.children:
            if isinstance(c, data.BoolLeaf):
                if value == c.getValue():
                    return value
        raise SemanticError(u"'%s' ist kein valider wert für '%s'"
                            % (value, nc.getName()))



class BoolChecker(ValueChecker):

    def convert(self, value, nc):
        value = value.strip()
        return len(value) and 1 or 0



TYPES = {
    data.IntLeaf: IntChecker(),
    data.NumericLeaf: NumericChecker(),
    data.DateLeaf: DateChecker(),
    data.TextLeaf: TextChecker(),
    data.TextAreaLeaf: TextAreaChecker(),
    data.ChoiceNode: ChoiceChecker(),
    data.BoolLeaf: BoolChecker(),
    data.RadioNode: RadioChecker()
}


def checkAndConvert(nc, value):
    try:
        checker = TYPES[nc.__class__]
    except KeyError:
        raise SemanticError("no type check found for '%s'" % str(nc.__class__))
    return checker.checkAndConvert(value, nc)

def check(nc, value):
    """Perform the semantic checks defined by nc on value.
    The nc parameter should be the meta-Node for the value.  The value
    must already be and instance of the native python type for the node.
    Raises SemanticError if the semantic checks fail.
    """
    try:
        checker = TYPES[nc.__class__]
    except KeyError:
        raise SemanticError("no type check found for '%s'" % str(nc.__class__))
    return checker.check(value, nc)


# vim:set ts=4 sw=4 si et sta sts=4:
