# -*- coding: utf-8 -*-

import logging
import re

from pylons import g, c
from mpulsweb.lib.translation import _

from formed.instance.backends.postgres import DBFactory as InstanceFactory
from formed.instance.backends.common import RuleEvalContext
from formed.instance.tree import PageInstanceNode
from formed.meta.data import PageNode

from mpulsweb.lib.helpers import tag, url_for, literal
from mpulsweb.lib.helper.case import walk_it 
from mpulsweb.lib.db import PostgresDBInterface

# name of the dbfield for phase
PHASEFIELD = "phase"
# name of flags
REQUIRED = re.compile('required')
PHASEDATE = re.compile('time')

# CHANGEMODES
# Can be configured json config
PC_MANUAL = '0'
PC_HALFAUTO = '1'
PC_FULLAUTO = '2'

log = logging.getLogger(__name__)

def get_flags(flags):
    result = {'required': [], 'time': []}
    if not flags: return result
    ft = flags.split(';')
    for f in ft:
        for p in f.split(':')[1].split(','):
            id = p.split('_')[1]
            if REQUIRED.match(f):
                result['required'].append(id)
            if PHASEDATE.match(f):
                result['time'].append(id)
    return result


def get_phase_successors(p, phaselist=None):
    suc = []
    if phaselist is None:
        phaselist = PhaseFactory().build()
    for phase in phaselist:
        if p == phase.id:
            suc.append(phase.start.id)
            suc.append(phase.end.id)
            for ppid in phase.end.suc:
                pp = phaselist.get_phase(ppid)
                suc.extend(get_phase_successors(pp.phase, phaselist))
    return suc


class ConsistenceCheckException(Exception):

    def __init__(self, value):
        self.value = value


class PhaseFactory:

    def __init__(self):
        self.phases = PhaseList()
        self.suc = {}
        self.pre = {}

        def _get_suc(p):
            ps = []
            s = g.mpuls_config.get('phases', 'successors')[0].get(p)
            for p in s:
                for pp in g.mpuls_config.get('phases', 'pairs')[0].get(p)[0]:
                    ps.append(pp[0])
            return ps 

        # Build successors and predessesors of the phases
        for p in g.mpuls_config.get('phases', 'phases'):
            pp = g.mpuls_config.get('phases', 'pairs')[0].get(p)
            if len(pp) == 2:
                self.suc[pp[0]] = []
                self.suc[pp[0]].extend(pp[1])
                self.suc[pp[1]] = []
                self.suc[pp[1]].extend(_get_suc(p))

        for k, v in self.suc.iteritems():
            for p in v:
                self.pre[p] = []
                self.pre[p].append(k)

    def build(self):
        '''Build phaselist from metadata'''
        for p in g.mpuls_config.get('phases', 'phases'):
            desc = g.mpuls_config.get('phases', 'description')[0].get(p)
            pair = g.mpuls_config.get('phases', 'pairs')[0].get(p)
            start = StartPhasePart(pair[0], p, 
                                   self.suc.get(pair[0], []), 
                                   self.pre.get(pair[0], []))
            end = EndPhasePart(pair[1], p,
                                   self.suc.get(pair[1], []), 
                                   self.pre.get(pair[1], []))
            self.phases.append(Phase(p, start, end, desc))
        self.phases.set_suc(self.suc)
        self.phases.set_pre(self.pre)
        return self.phases

    def load(self, id):
        '''Build phaselist from metadata and load current state from instance tree'''
        self.build()
        factory = InstanceFactory(g.formedTree, PostgresDBInterface())
        it = factory.loadInstanceTreeByIdentifier(id)
        for p in self.phases:
            start = p.start
            end = p.end
            start.load(it)
            end.load(it)
        return self.phases

class PhaseList(list):

    def set_suc(self, suc):
        self.suc = suc

    def set_pre(self, pre):
        self.pre = pre 

    def get_current_phase_id(self):
        '''Returns id of current active phasepart'''
        p = self.get_current_phase()
        return p.id

    def get_current_phase(self):
        '''Returns current active phasepart'''
        for phase in self:
            p1 = phase.getStart()
            if p1.is_active():
                return p1
            p2 = phase.getEnd()
            if p2.is_active():
                return p2

        # Return dummy phase for unknown phase
        p = EndPhasePart('-1', None, ['1'], [])
        p.phase = '-1'
        return p

    def get_phase(self, id):
        for p in self:
            if id == p.start.id: return p.start
            if id == p.end.id: return p.end

        # Return dummy phase for unknown phase
        if id == '-1':
            p = EndPhasePart('-1', None, ['1'], [])
            p.phase = '-1'
            return p

    def get_max_phase_id(self):
        return max(self.get_reachable_path())

    def get_reachable_path(self, phasepart_id='0', only_reachable=True):
        '''Returns id of max reachable phase which is valid starting from
        phasepart_id'''

        path = ['-1'] # default unknown phase
        if not self.get_phase(phasepart_id).is_ok() and only_reachable:
            return path
        for pp in self.suc.get(phasepart_id, []):
            if not self.get_phase(pp).is_ok() and only_reachable: 
                continue
            path.extend(self.get_reachable_path(pp, only_reachable))
            path.append(pp)
        path.append(phasepart_id)
        return path

    def get_path(self, phasepart_id, only_reachable=False):
        '''Returns a list of phaseparts id from start to the given phasepart_id'''
        path = []
        for pp in self.pre.get(phasepart_id, []):
            # Ignore phase which are not reachable
            if not self.get_phase(pp).is_ok() and only_reachable: 
                continue
            path.extend(self.get_path(pp))
            path.append(pp)
        path.append(phasepart_id)
        return path

    def is_on_path(self, findid, pathend, only_reachable=False):
        findid = str(findid)
        return findid in self.get_path(pathend, only_reachable)

    def is_valid(self):
        '''Returns true if all phaseparts up to the current phase are ok.
        This means all requierd fields are filled out and all checks for
        rules are ok. Else return False
        '''
        try:
            self.check_consistence()
            return True
        except ConsistenceCheckException, e:
            return False

    def check_consistence(self, fields=None, phase=None):
        CASE_NOT_CONSISTENT = _("The phase '%s' is in a non consitent \
                                state and has missing required fields: %s")
        if phase is not None:
            current_phase = phase
        else:
            current_phase = self.get_current_phase_id()
        path = self.get_path(str(current_phase))
        for phases in self:
            p1, p2 = phases.getStart(), phases.getEnd()
            if int(p1.id) <= int(current_phase) and p1.id in path:
                if not p1.is_ok(fields, check_setting_phase=False):
                    errors = []
                    errors.append('<ol>')
                    for e in p1.errors:
                        errors.append('<li>%s</li>' % e)
                    errors.append('</ol>')
                    raise ConsistenceCheckException(CASE_NOT_CONSISTENT
                                                    % (phases.description, "".join(errors)))
            
            if int(p2.id) <= int(current_phase) and p2.id in path:
                print p2.id 
                if not p2.is_ok(fields, check_setting_phase=False):
                    errors = []
                    errors.append('<ol>')
                    for e in p2.errors:
                        errors.append('<li>%s</li>' % e)
                    errors.append('</ol>')
                    raise ConsistenceCheckException(CASE_NOT_CONSISTENT
                                                    % (phases.description, "".join(errors)))

class Phase:

    def __init__(self, id, start, end, description):
        self.id = id 
        self.start = start
        self.end = end
        self.description = description

    def getStart(self):
        return self.start

    def getEnd(self):
        return self.end

    def getDuration(self):
        t1 = self.start.getTime()
        t2 = self.end.getTime()
        if not t1 or not t2:
            return None
        return t2 - t1

    def getDescription(self):
        return literal(self.description)

    def isRunning(self):
        return self.start.is_active()

class PhasePart:

    def __init__(self, id, phase, suc, pre):
        self.id = id
        self.it = None 
        self.phase = phase
        self.active = False
        self.date = None
        self.datefield = None
        self.suc = suc
        self.pre = pre
        self.fields = []
        self.errors = []

        # If id is 0 (which should be the start phasepart of the first phase,
        # add "unknown" phase as predesessor) 
        if self.id == '0':
            self.pre.append('-1')

        # Get phase relevat dates
        for p in g.formedTree.walk():
            flags = get_flags(p.getFlags())
            if self.id in flags['time']:
                self.datefield = p.getName()

        if self.datefield is None:
            log.warning("Phase %s has no associated datefield", id)

    def getDescription(self):
        desc = []
        desc.append(g.mpuls_config.get('phases',
                                       'description')[0].get(self.phase))
        if self.phase != '-1':
            if isinstance(self, StartPhasePart):
                desc.append(_('Running'))
            else:
                desc.append(_('Finished'))
        return " ".join(desc)

    def load(self, it):
        formed = it.getFormEd()
        self.it = it
        # Check if the current phasepart is active
        case_id = it.getRootNode().getIdentifier()
        realid = "%s:%s" % (PHASEFIELD, case_id)
        field = it.getItem(realid)
        self.active = field.getValue() == int(self.id)

        for p in walk_it(it.getRootNode()):
            meta = p.getMeta()
            if isinstance(meta, PageNode):
                page = meta.getName()
            flags = get_flags(meta.getFlags())

            name = meta.getName()
            identifier = p.getIdentifier()
            # Check if this fields is a phase relvant required field 
            if self.id in flags['required']:
                # Only add warnigs and rules with "required" in der name
                rules = [rule for rule in formed.name2markingrules.get(name, ()) 
                         if rule.isWarning() and REQUIRED.match(rule.getName())]
                self.fields.append(Field(identifier, page, rules, it))

        # Load phase relvant datefield
        if self.datefield is not None:
            realid = "%s:%s" % (self.datefield, case_id)
            field = it.getItem(realid)
            warnings = [] 
            self.date = field.getValue()

    def is_neighbor(self, phase):
        return (str(phase )in self.suc) or (str(phase )in self.pre)

    def getLinks(self):
        pages = {}
        order = []
        p2 = {}
        if len(self.fields) > 0:
            for field in [field for field in self.fields]:
                if not p2.has_key(field.page):
                    p2[field.page] = []
                p2[field.page].append(field)
        for p, f in p2.iteritems():
            if p == 'ee': continue #(hidden)
            case_id = self.it.getRootNode().getIdentifier()
            realid = "%s:%s" % (p, case_id)
            item = self.it.getItem(realid)
            key = item.getMeta().getDescription()
            pages[key] = f
            order.append(key)
        order.sort()
        return pages, order

    def getTime(self):
        return self.date

    def is_active(self):
        return self.active

    def is_anonymizable(self):
        return self.id not in g.mpuls_config.get('phases', 'notanonymizable')

    def is_ok(self, fields=None, check_setting_phase=True):

        # 1. Simulate switching the phase
        # Setting values will trigger evaluation error rules. This way we can
        # check logic rules which are modeled as errors.
        errors = []
        if check_setting_phase and self.it:
            case_id = self.it.getRootNode().getIdentifier()
            realid = "%s:%s" % (PHASEFIELD, case_id)
            try:
                values = {realid: self.id}
                self.it.setData(values)
            except:
                return False
        

        # 2. Check rules
        for field in self.fields:
            if fields is None:
                # No fields provided -> checkall
                if not field.is_ok():
                    errors.append(field.label)
            elif field.name in fields:
                if not field.is_ok():
                    errors.append(field.label)
        self.errors = errors
        return len(errors) == 0


class StartPhasePart(PhasePart):
    pass

class EndPhasePart(PhasePart):
    pass

class Field:

    def __init__(self, id, page, rules, it):
        self.id = id
        self.page = page
        self.it = it
        self.label = None
        self.name = None
        self.conditions = []

        field = self.it.getItem(self.id)
        meta = field.getMeta()
        self.name = meta.getName()
        self.label = meta.getDescription()
        for rule in rules:
            self.conditions.append(Condition(field, rule, it))

    def get_url(self):
        element_id = ":".join(self.id.split(':')[1:])
        return  url_for(controller='/formularpage', action='select',
                        id=('%s:%s' % (self.page, element_id)),
                        anchor=('%s:%s' % (self.name,
                        element_id)))
    def get_link(self):
        out = []
        if self.is_ok():
            style="ok"
            atitle=_('Field is OK and requires no further editing')
        else:
            style="error"
            atitle=_('Field is not OK and requires further editing')
        out.append(tag("a", class_=style,
                       href=self.get_url(),
                       title_=atitle))
        out.append('%s' % self.label)
        out.append('</a>')
        return "".join(out)

    def has_relevant_rules(self):
        '''Returns true if the field has relevant rules. Else False e.g if the
        required field is in a condition which is not true'''
        for c in self.conditions:
            if c.is_relevant():
                return True
        return False

    def is_ok(self):
        for c in self.conditions:
            if not c.is_ok():
                return False
        return True


class Condition:

    def __init__(self, field, rule, it):
        self.field = field
        self.rule= rule 
        self.it = it 

    def is_ok(self):
        '''Returns True if expr is valid for the current value in field
        identified with name'''
        return self.evaluate()

    def is_relevant(self):
        id = self.field.getIdentifier()
        context = RuleEvalContext(self.it, id)
        return self.it.rule_available(self.rule, context)

    def evaluate(self):
        id = self.field.getIdentifier()
        context = RuleEvalContext(self.it, id)
        if not self.is_relevant(): return True
        for result, variables in self.rule.getExpr().evaluate_instances(context):
            if not result and id in variables:
                return False
            return True

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
